"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

# drift detection and all
# run.log_dataset(
#     dataset_name='train',
#     features=X_train,
#     predictions=y_pred_train,
#     actuals=y_train,
# )
# run.log_dataset(
#     dataset_name='test',
#     features=X_test,
#     predictions=y_pred_test,
#     actuals=y_test,
# )

import re
import json
from git import Repo
from typing import Union, Dict, Any

from nbox.utils import logger
from nbox.auth import secret, AuthConfig
from nbox.init import nbox_ws_v1
from nbox.lmao.lmao_rpc_client import (
  LMAO_Stub, # main stub class
  Record,
  ListProjectsRequest,
  Project as ProjectProto,
)
from nbox.hyperloop.common.common_pb2 import Resource
from nbox import messages as mpb

"""
functional components of LMAO
"""

def get_lmao_stub() -> LMAO_Stub:
  # lmao_stub = LMAO_Stub(url = "http://localhost:8080/monitoring", session = nbox_ws_v1._session)
  lmao_stub = LMAO_Stub(url = secret(AuthConfig.url) + "/monitoring", session = nbox_ws_v1._session)
  return lmao_stub

def get_record(k: str, v: Union[int, float, str]) -> Record:
  """Function to create a Record protobuf object from a key and value."""
  _tv = type(v)
  assert _tv in [int, float, str], f"[key = {k}] '{_tv}' is not a valid type"
  _vt = {
    int: Record.DataType.INTEGER,
    float: Record.DataType.FLOAT,
    str: Record.DataType.STRING,
  }[_tv]
  record = Record(key = k, value_type = _vt)
  if _tv == int:
    record.integer_data.append(v)
  elif _tv == float:
    record.float_data.append(v)
  elif _tv == str:
    record.string_data.append(v)
  return record

def get_git_details(folder):
  """If there is a `.git` folder in the folder, return some details for that."""
  repo = Repo(folder)

  # check for any unstaged files
  uncommited_files = {}
  diff = repo.index.diff(None)
  for f in diff:
    path = f.a_path or f.b_path # when new file is added, a_path is None
    uncommited_files[path] = f.change_type
  if uncommited_files:
    logger.warning(f"Uncommited files: {uncommited_files}")

  # get the remote url
  try:
    remote_url = repo.remote().url

    # clean the remote_url because it can sometimes contain the repo token as well.
    # this can become security hazard. so if you have an example, that is not suppported:
    #   go ahead, make a PR!
    if "github.com" in remote_url:
      remote_url = re.sub(r"ghp_\w+@", "", remote_url)
  except ValueError:
    remote_url = None

  # get the size of the repository
  size = None
  for line in repo.git.count_objects("-v").splitlines():
    if line.startswith("size:"):
      size = int(line[len("size:") :].strip())
  if size > (1 << 30):
    logger.warning(f"Repository size over 1GB, you might want to work on it")

  return {
    "remote_url": remote_url,
    "branch": repo.active_branch.name,
    "commit": repo.head.commit.hexsha,
    "uncommited_files": uncommited_files,
    "untracked_files": repo.untracked_files,
    "size": size,
  }

def get_project(project_id: str) -> ProjectProto:
  lmao_stub = get_lmao_stub()
  project: ProjectProto = lmao_stub.get_project(ListProjectsRequest(
    workspace_id = secret(AuthConfig.workspace_id),
    project_id_or_name = project_id
  ))
  return project


def resource_from_dict(d: Dict[str, Any]):
  return Resource(
    cpu = str(d["cpu"]),
    memory = str(d["memory"]),
    disk_size = str(d["disk_size"]),
    gpu = str(d["gpu"]),
    gpu_count = str(d["gpu_count"]),
    timeout = int(d["timeout"]),
    max_retries = int(d["max_retries"]),
  )

"""
Common structs
"""


class ExperimentConfig:
  def __init__(
    self,
    run_kwargs: Dict[str, Any],
    git: Dict[str, Any],
    resource: Resource,
    cli_comm: str,
    save_to_relic: bool = True,
    enable_system_monitoring: bool = False,
  ):
    """In an ideal world this would be a protobuf message, but we are not there yet. This contains all the
    things that are stored in the DB and FE can use this to render elements in the UI.

    Args:
      run_kwargs (Dict[str, Any]): All the arguments that the user passed to the `nbx lmao run ...` CLI
      git (Dict[str, Any]): details for the git repo
      resource (Resource): Resource pb object that contains the details of the resource
      cli_comm (str): The CLI command that was used to run the experiment, user can use this to reproduce the experiment
      save_to_relic (bool): If the user wants to save the experiment to the relic by default
      enable_system_monitoring (bool): If the user wants to enable system monitoring
    """
    self.run_kwargs = run_kwargs
    self.git = git
    self.resource = resource
    self.cli_comm = cli_comm
    self.save_to_relic = save_to_relic
    self.enable_system_monitoring = enable_system_monitoring

  def to_dict(self):
    return {
      "run_kwargs": self.run_kwargs,
      "git": self.git,
      "resource": mpb.message_to_dict(self.resource),
      "cli_comm": self.cli_comm,
      "save_to_relic": self.save_to_relic,
      "enable_system_monitoring": self.enable_system_monitoring,
    }

  def to_json(self):
    return json.dumps(self.to_dict())

  @classmethod
  def from_json(cls, json_str):
    d = json.loads(json_str)
    d["resource"] = resource_from_dict(d["resource"])
    return cls(**d)



class LiveConfig:
  def __init__(
    self,
    resource: Resource,
    cli_comm: str,
    enable_system_monitoring: bool = False,
    extra_kwargs: Dict[str, Any] = {},
  ):
    self.resource = resource
    self.cli_comm = cli_comm
    self.enable_system_monitoring = enable_system_monitoring
    self.extra_kwargs = extra_kwargs

  def to_dict(self):
    return {
      "resource": mpb.message_to_dict(self.resource),
      "cli_comm": self.cli_comm,
      "enable_system_monitoring": self.enable_system_monitoring,
      "extra_kwargs": self.extra_kwargs,
    }
  
  def to_json(self):
    return json.dumps(self.to_dict())
  
  @classmethod
  def from_json(cls, json_str) -> 'LiveConfig':
    d = json.loads(json_str)
    d["resource"] = resource_from_dict(d["resource"])
    _cls = cls(**d)
    return _cls


"""
Constants
"""

# do not change these it can become a huge pain later on
LMAO_RELIC_NAME = "experiments"
LMAO_RM_PREFIX = "NBXLmao-"
LMAO_SERVING_FILE = "NBXLmaoServingCfg.json"
LMAO_ENV_VAR_PREFIX = "NBX_LMAO_"



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for measuring elapsed time. From TensorBoard (Apache-v2)."""

import contextlib
import logging
import threading
import time

from nbox.utils import logger


def log_latency(region_name_or_function_to_decorate, log_level=None):
  """Log latency in a function or region.
  Three usages are supported. As a decorator:

  ```python
  >>> @log_latency
  ... def function_1():
  ...   pass
  ...
  ```
  
  As a decorator with a custom label for the region:
  ```python
  >>> @log_latency("custom_label")
  ... def function_2():
  ...   pass
  ...
  ```
  
  As a context manager:
  ```python
  >>> def function_3():
  ...   with log_latency("region_within_function"):
  ...     pass
  ...
  ```
  
  - region_name_or_function_to_decorate: Either: a `str`, in which
    case the result of this function may be used as either a
    decorator or a context manager; or a callable, in which case
    the result of this function is a decorated version of that
    callable.
  - log_level: Optional integer logging level constant. Defaults to
    `logging.INFO`.

  Returns:
    A decorated version of the input callable, or a dual
    decorator/context manager with the input region name.
  """

  if log_level is None:
    log_level = logging.DEBUG

  if isinstance(region_name_or_function_to_decorate, str):
    region_name = region_name_or_function_to_decorate
    return _log_latency(region_name, log_level)
  else:
    function_to_decorate = region_name_or_function_to_decorate
    qualname = getattr(function_to_decorate, "__qualname__", None)
    if qualname is None:
      qualname = str(function_to_decorate)
    decorator = _log_latency(qualname, log_level)
    return decorator(function_to_decorate)


class _ThreadLocalStore(threading.local):
  def __init__(self):
    self.nesting_level = 0


_store = _ThreadLocalStore()


@contextlib.contextmanager
def _log_latency(name, log_level):
  if not logger.isEnabledFor(log_level):
    yield
    return

  start_level = _store.nesting_level
  try:
    started = time.time()
    _store.nesting_level = start_level + 1
    thread = threading.current_thread()
    msg = f"ENTER '{name}'"
    if logger.getEffectiveLevel() == 10:
      indent = (" " * 2) * start_level
      msg = f"{thread.name} [{thread.ident}]{indent} {msg}"
    _log(log_level, msg)
    yield
  finally:
    _store.nesting_level = start_level
    elapsed = time.time() - started
    msg = f"LEAVE '{name}' - {elapsed:0.6f}s elapsed"
    if logger.getEffectiveLevel() == 10:
      indent = (" " * 2) * start_level
      msg = f"{thread.name} [{thread.ident}]{indent} {msg}"
    _log(log_level, msg)


def _log(log_level, msg, *args):
  # Forwarding method to ensure that all logging statements
  # originating in this module have the same line number; if the
  # "ENTER" log is on a line with 2-digit number and the "LEAVE" log
  # is on a line with 3-digit number, the logs are misaligned and
  # harder to read.
  logger.log(log_level, msg, *args)

if __name__ == "__main__":
  @log_latency("some fancy function")
  def run(x = 2):
    print("run")
    time.sleep(1)
    return x
  run()




import os
import tabulate
from copy import deepcopy
from functools import lru_cache

from nbox.auth import secret, AuthConfig
from nbox.init import nbox_ws_v1
from nbox.utils import logger, get_mime_type
from nbox.relics.proto.relics_rpc_pb2 import ListRelicsRequest
from nbox.relics.proto.relics_pb2 import RelicFile
from nbox.relics.relics_rpc_client import RelicStore_Stub

def get_relic_file(fpath: str, username: str, workspace_id: str = ""):
  workspace_id = workspace_id or secret(AuthConfig.workspace_id)
  # assert os.path.exists(fpath), f"File {fpath} does not exist"
  # assert os.path.isfile(fpath), f"File {fpath} is not a file"

  # clean up fpath, remove any trailing slashes
  # trim any . or / from prefix and suffix
  fpath_cleaned = fpath.strip("./")

  extra = {}
  if os.path.exists(fpath):
    file_stat = os.stat(fpath)
    extra = {
      "created_on": int(file_stat.st_mtime),    # int
      "last_modified": int(file_stat.st_mtime), # int
      "size": max(1, file_stat.st_size),        # bytes
    }
  content_type = get_mime_type(fpath_cleaned, "application/octet-stream")
  return RelicFile(
    name = fpath_cleaned,
    username = username,
    type = RelicFile.RelicType.FILE,
    workspace_id = workspace_id,
    content_type=content_type,
    **extra
  )


@lru_cache()
def get_relics_stub():
  # url = "http://0.0.0.0:8081/relics" # debug mode
  url = secret("nbx_url") + "/relics"
  logger.debug("Connecting to RelicStore at: " + url)
  session = deepcopy(nbox_ws_v1._session)
  stub = RelicStore_Stub(url, session)
  return stub


def print_relics(workspace_id: str = ""):
  stub = get_relics_stub()
  workspace_id = workspace_id or secret(AuthConfig.workspace_id)
  req = ListRelicsRequest(workspace_id = workspace_id,)
  out = stub.list_relics(req)
  headers = ["relic_id", "relic_name",]
  rows = [[r.id, r.name,] for r in out.relics]
  for l in tabulate.tabulate(rows, headers).splitlines():
    logger.info(l)



from nbox.relics.relics_rpc_client import RelicStore_Stub
from nbox.relics.proto import relics_pb2, relics_rpc_pb2, common_pb2
from nbox.relics.client import Relics, UserAgentType
from nbox.relics.utils import get_relic_file, get_relics_stub, print_relics


"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

import os
from json import dumps, loads
from functools import lru_cache
from typing import Dict, Any, List, Optional, Union

import nbox.utils as U
from nbox.utils import logger, SimplerTimes
from nbox.auth import secret, AuthConfig, JobDetails
from nbox.relics import Relics

# all the sublime -> hyperloop stuff
from nbox.lmao.lmao_rpc_client import (
  AgentDetails,
  RunLog,
  Run as RunProto,
  InitRunRequest
)
from nbox.observability.system import SystemMetricsLogger
from nbox.lmao.common import get_git_details, get_lmao_stub, get_record, get_project



"""
Client library that the user will use to interact with the LMAO server.
"""

class _lmaoConfig:
  # _lmaoConfig.kv contains all the objects that the class LMAO needs to work correctly, however
  # we will also need to take care of the things we want to show 
  kv = {}
  def set(
    project_id: str,
    experiment_id: str,
    save_to_relic: bool,
    enable_system_monitoring: bool,
    store_git_details: bool,
  ) -> None:
    _lmaoConfig.kv = {
      "project_id": project_id,
      "experiment_id": experiment_id,
      "save_to_relic": save_to_relic,
      "enable_system_monitoring": enable_system_monitoring,
      "store_git_details": store_git_details,
    }


class LmaoTypes:
  EXP = "experiment"
  LIVE = "live"

  def all() -> List[str]:
    return [LmaoTypes.EXP, LmaoTypes.LIVE]


class Lmao():
  def __init__(
    self,
    project_id: Optional[str] = "",
    experiment_id: Optional[str] = "",
    metadata: Dict[str, Any] = {},
    save_to_relic: bool = True,
    enable_system_monitoring: bool = False,
    store_git_details: bool = True,
  ) -> None:
    """`Lmao` is the client library for using NimbleBox Monitoring. It talks to your monitoring instance running on your build
    and stores the information in the `project_name` or `project_id`. This object inherently doesn't care what you are actually
    logging and rather concerns itself with ensuring storage.

    **Note**: All arguments are optional, if the `_lmaoConfig.kv` is set.

    Args:
      project_name (str, optional): The name of the project. Defaults to "".
      project_id (str, optional): The id of the project. Defaults to "".
      experiment_id (str, optional): The id of the experiment. Defaults to "".
      metadata (Dict[str, Any], optional): Any metadata that you want to store. Defaults to {}.
      save_to_relic (bool, optional): Whether to save the data to the relic. Defaults to False.
      enable_system_monitoring (bool, optional): Whether to enable system monitoring. Defaults to False.
      store_git_details (bool, optional): Whether to store git details. Defaults to True.
    """
    if _lmaoConfig.kv:
      # load all the values from the config
      project_id = _lmaoConfig.kv["project_id"]
      experiment_id = _lmaoConfig.kv["experiment_id"]
      save_to_relic = _lmaoConfig.kv["save_to_relic"]
      enable_system_monitoring = _lmaoConfig.kv["enable_system_monitoring"]
      store_git_details = _lmaoConfig.kv["store_git_details"]

    self.project_id = project_id
    self.experiment_id = experiment_id
    self.save_to_relic = save_to_relic
    self.enable_system_monitoring = enable_system_monitoring
    self.store_git_details = store_git_details
    self.workspace_id = secret(AuthConfig.workspace_id)

    self.agent = secret.get_agent_details()

    # now set the supporting keys
    self._total_logged_elements = 0
    self.completed = False
    self.relic: Relics = None
    self.sys_montior: SystemMetricsLogger = None

    # create connection and initialise the run
    self.lmao = get_lmao_stub()
    self.project = get_project(project_id = self.project_id)
    if self.project is None:
      raise Exception(f"Project with id {self.project_id} does not exist")

    self.config = self._get_config(metadata = metadata)
    self.run: RunProto = self._init_experiment(
      project_id = project_id,
      config = self.config,
      experiment_id = experiment_id
    )

    if self.save_to_relic:
      self.relic = Relics(id = self.project.relic_id, prefix = self.run.save_location)
      logger.info(f"Will store everything in relic '{self.relic}'")

    # system metrics monitoring, by default is enabled optionally turn it off
    if self.enable_system_monitoring:
      raise NotImplementedError
      self.sys_montior = SystemMetricsLogger(self)
      self.sys_montior.start()

  def __repr__(self) -> str:
    return f"Lmao({self.project_id}, {self.experiment_id}, {self._total_logged_elements})"

  def __del__(self):
    if self.sys_montior is not None:
      self.sys_montior.stop()

  def _get_config(self, metadata: Dict[str, Any]):
    # this is the config value that is used to store data on the plaform, user cannot be allowed to have
    # like a full access to config values
    log_config: Dict[str, Any] = {
      "run_kwargs": metadata
    }

    # check if the current folder from where this code is being executed has a .git folder
    # NOTE: in case of NBX-Jobs the current folder ("./") is expected to contain git by default
    log_config["git"] = None
    if os.path.exists(".git") and self.store_git_details:
      log_config["git"] = get_git_details("./")
    return log_config

  def _init_experiment(self, project_id, config: Dict[str, Any] = {}, experiment_id: str = "") -> RunProto:
    # update the server or create new experiment
    agent_details = AgentDetails(
      workspace_id = self.workspace_id,
      nbx_job_id = "local",
      nbx_run_id = U.SimplerTimes.get_now_str(),
    )
    if type(self.agent) == JobDetails:
      a: JobDetails = self.agent
      agent_details.type = AgentDetails.NBX.JOB
      agent_details.nbx_job_id = a.job_id
      agent_details.nbx_run_id = a.run_id

    if experiment_id:
      action = "Updated"
      run_details = self.lmao.get_run_details(Run(
        workspace_id = self.workspace_id,
        project_id = project_id,
        experiment_id = experiment_id,
      ))
      if not run_details:
        # TODO: Make a custom exception of this
        raise Exception("Server Side exception has occurred, Check the log for details")
      if run_details.experiment_id:
        # means that this run already exists so we need to make an update call
        ack = self.lmao.update_run_status(Run(
          workspace_id = self.workspace_id,
          project_id = project_id,
          experiment_id = run_details.experiment_id,
          agent = agent_details,
          update_keys = ["agent"],
        ))
        if not ack.success:
          raise Exception(f"Failed to update run status! {ack.message}")
    else:
      action = "Created"
      run_details = self.lmao.init_run(
        InitRunRequest(
          workspace_id = self.workspace_id,
          project_id = project_id,
          agent_details=agent_details,
          config = dumps(config),
        )
      )

    logger.info(
      f"{action} experiment tracker\n"
      f" project: {project_id}\n"
      f"      id: {run_details.experiment_id}\n"
      f"    link: {secret(AuthConfig.url)}/workspace/{self.workspace_id}/projects/{project_id}#Experiments\n"
    )

    return run_details

  """The functions below are the ones supposed to be used."""

  @property
  def run_config(self) -> Dict[str, Any]:
    return loads(self.run.config)

  @lru_cache(maxsize=1)
  def get_relic(self):
    """Get the underlying Relic for more advanced usage patterns."""
    if self.save_to_relic:
      return self.relic
    raise Exception("Relic is not enabled for this run, set save_to_relic = True")

  def log(self, y: Dict[str, Union[int, float, str]], step = None, *, log_type: str = RunLog.LogType.USER):
    """Log a single level dictionary to the platform at any given step. This function does not really care about the
    information that is being logged, it just logs it to the platform."""
    if self.completed:
      raise Exception("Run already completed, cannot log more data!")

    step = step or self._total_logged_elements
    if step < 0:
      raise Exception("Step must be >= 0")
    run_log = RunLog(
      workspace_id = self.workspace_id,
      project_id=self.project_id,
      experiment_id = self.run.experiment_id,
      log_type=log_type
    )
    for k,v in y.items():
      # TODO: @yashbonde replace Record with RecordColumn
      record = get_record(k, v)
      record.step = step
      run_log.data.append(record)

    ack = self.lmao.on_log(run_log)
    if not ack.success:
      logger.error(f"  >> Server Error\n{ack.message}")
      raise Exception("Server Error")

    self._total_logged_elements += 1

  def end(self):
    """End the run to declare it complete. This is more of a convinience function than anything else. For example when you
    are monitoring a live API you may never know when the experiment is complete. This function locks the experiment name
    so it can't be modified."""
    if self.completed:
      logger.error("Run already completed, cannot end it again!")
      return None

    logger.info("Ending run")
    ack = self.lmao.on_train_end(self.run)
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")
    self.completed = True
    if self.enable_system_monitoring:
      self.sys_montior.stop()

  def save_file(self, *files: List[str]):
    logger.info(f"Saving files: {files}")
    return self.add_files(*files)

  def add_files(self, *files: List[str]):
    """
    Register a file save. If `save_to_relics` is not set, this function is a no-op.

    Args:
      files: The list of files to save. This can be a list of files or a list of folders. If a folder is passed, all the files in the folder will be uploaded.
    """
    # manage all the complexity of getting the list of RelicFile
    all_files = []
    for folder_or_file in files:
      if os.path.isfile(folder_or_file):
        all_files.append(folder_or_file)
      elif os.path.isdir(folder_or_file):
        all_files.extend(U.get_files_in_folder(folder_or_file, abs_path=False))
      else:
        raise Exception(f"File or Folder not found: {folder_or_file}")

    # when we add something cool we can use this
    logger.debug(f"Storing {len(all_files)} files")
    if self.save_to_relic:
      relic = self.get_relic()
      for f in all_files:
        relic.put(f)
    return all_files



"""
Contains our custom logger and formatters so you can log things like a pro.
It can generate both inline logs as well as JSON logs (easier to consume by other tools).
Json logs go really well with our custom logger for jobs and deploy where you can
logs arbitrary objects and have them nicely formatted in the console.

Parts of this code are taken from the awesome `pyjsonlogger` library
([Pypi](https://pypi.org/project/jsonformatter/))
"""
# BSD 2-Clause License
# 
# Copyright (c) 2019, MyColorfulDays
# Copyright (c) 2023, NimbleBox.ai
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#  
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
import logging
import traceback
from inspect import istraceback
from datetime import date, datetime, time, timezone


def get_logger(env):
  logger = logging.getLogger("utils")
  lvl = env.NBOX_LOG_LEVEL("info").upper()
  logger.setLevel(getattr(logging, lvl))

  if env.NBOX_JSON_LOG(False):
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(JsonFormatter())
    logger.addHandler(logHandler)
  else:
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(InlineFormatter(
      '[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d] %(message)s',
      datefmt = "%Y-%m-%dT%H:%M:%S%z"
    ))
    logger.addHandler(logHandler)

  return logger


##############

class _LogObject:
  def __init__(self, log, msg):
    self.log = log
    self.msg = msg

def lo(msg, /, *args, **kwargs):
  if args:
    msg += " " + " ".join([str(a) for a in args])
  return _LogObject(kwargs, msg)

##############

class InlineFormatter(logging.Formatter):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def format(self, record):
    if type(record.msg) == _LogObject:
      _l: _LogObject = record.msg
      if _l.log:
        record.msg = f"{_l.msg}\n" + json_dumps_pretty(_l.log)
      else:
        record.msg = _l.msg
    return super().format(record)

class JsonFormatter(logging.Formatter):
  def format(self, record):
    """Formats a log record and serializes to json"""

    log_record = {
      "timestamp": datetime.fromtimestamp(record.created, tz=timezone.utc).isoformat(),
      "levelname": record.levelname,
      "logtype": "nbox_jlv1",
      "log": {
        "code": {
          "file": record.filename,
          "line": record.lineno,
        },
      }
    }

    if isinstance(record.msg, dict):
      log_record["message"] = "click to see data"
      log_record["log"]["data"] = record.msg
    elif isinstance(record.msg, _LogObject):
      _lo: _LogObject = record.msg
      log_record["message"] = _lo.msg
      if _lo.log:
        log_record["log"]["data"] = _lo.log
    elif record.msg:
      lines = record.getMessage().splitlines()
      if len(lines) == 1:
        # you only have one line, don't bother to create lot of data
        log_record["message"] = lines[0]
      else:
        log_record["message"] = lines[0][:10] + "..."
        log_record["log"]["lines"] = lines

    json_str = json_dumps_tight(log_record)
    return json_str

##############

class _JsonEncoder(json.JSONEncoder):
  """
  A custom encoder extending the default JSONEncoder
  """
  def default(self, obj):
    if isinstance(obj, (date, datetime, time)):
      return self.format_datetime_obj(obj)
    elif istraceback(obj):
      return ''.join(traceback.format_tb(obj)).strip()
    elif type(obj) == Exception \
        or isinstance(obj, Exception) \
        or type(obj) == type:
      return str(obj)
    try:
      return super(_JsonEncoder, self).default(obj)
    except TypeError:
      try:
        return str(obj)
      except Exception:
        return None

  def format_datetime_obj(self, obj):
    return obj.isoformat()


json_dumps_tight = lambda log_record: json.dumps(
  log_record,
  cls = _JsonEncoder,
  ensure_ascii = True,
  separators = (',', ':'),
)

json_dumps_pretty = lambda log_record: json.dumps(
  log_record,
  cls = _JsonEncoder,
  ensure_ascii = True,
  indent = 2,
  separators = (',', ': '),
)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/relics.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12proto/relics.proto\"Y\n\x0b\x42\x61\x63kendInfo\x12\x19\n\x07\x62\x61\x63kend\x18\x01 \x01(\x0e\x32\x08.Backend\x12\x17\n\x0fnbx_resource_id\x18\x02 \x01(\t\x12\x16\n\x0enbx_access_key\x18\x03 \x01(\t\"\xfd\x01\n\x05Relic\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\ncreated_on\x18\x03 \x01(\x03\x12\x15\n\rlast_modified\x18\x04 \x01(\x03\x12\x0f\n\x07starred\x18\x05 \x01(\x08\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x14\n\x0cworkspace_id\x18\x07 \x01(\t\x12\x12\n\npermission\x18\x08 \x03(\t\x12\x10\n\x08ui_field\x18\t \x03(\t\x12\x12\n\ncreated_by\x18\n \x01(\t\x12\x1a\n\x04\x61uth\x18\x0b \x01(\x0b\x32\x0c.BackendInfo\x12$\n\x0b\x62ucket_meta\x18\x64 \x01(\x0b\x32\x0f.BucketMetadata\"\xb9\x01\n\x0e\x42ucketMetadata\x12\x13\n\x0b\x62ucket_name\x18\x01 \x01(\t\x12\x0e\n\x06region\x18\x02 \x01(\t\x12\x19\n\x07\x62\x61\x63kend\x18\x03 \x01(\x0e\x32\x08.Backend\x12\x34\n\x0b\x62ucket_tags\x18\x04 \x03(\x0b\x32\x1f.BucketMetadata.BucketTagsEntry\x1a\x31\n\x0f\x42ucketTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x8d\x04\n\tRelicFile\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ncreated_on\x18\x02 \x01(\x03\x12\x15\n\rlast_modified\x18\x03 \x01(\x03\x12\x0f\n\x07starred\x18\x04 \x01(\x08\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x0c\n\x04size\x18\x06 \x01(\x03\x12\x0f\n\x07\x63omment\x18\x07 \x01(\t\x12\x10\n\x08username\x18\x08 \x01(\t\x12\"\n\x04type\x18\t \x01(\x0e\x32\x14.RelicFile.RelicType\x12\x14\n\x0cworkspace_id\x18\n \x01(\t\x12\x12\n\nrelic_name\x18\x0b \x01(\t\x12\x10\n\x08relic_id\x18\x0c \x01(\t\x12\x14\n\x0c\x63ontent_type\x18\r \x01(\t\x12\x10\n\x08\x64ownload\x18\x0e \x01(\x08\x12\x0b\n\x03url\x18\x64 \x01(\t\x12(\n\x07headers\x18\x65 \x03(\x0b\x32\x17.RelicFile.HeadersEntry\x12\"\n\x04\x62ody\x18\x66 \x03(\x0b\x32\x14.RelicFile.BodyEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\tBodyEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"7\n\tRelicType\x12\t\n\x05UNSET\x10\x00\x12\x08\n\x04\x46ILE\x10\x01\x12\n\n\x06\x46OLDER\x10\x02\x12\t\n\x05RELIC\x10\x03\"=\n\nRelicFiles\x12\x14\n\x0cworkspace_id\x18\x01 \x01(\t\x12\x19\n\x05\x66iles\x18\x02 \x03(\x0b\x32\n.RelicFile*a\n\x07\x42\x61\x63kend\x12\t\n\x05UNSET\x10\x00\x12\x07\n\x03NBX\x10\x01\x12\n\n\x06\x41WS_S3\x10\x02\x12\x0b\n\x07GCP_GCS\x10\x03\x12\x0e\n\nAZURE_BLOB\x10\x04\x12\n\n\x06OCI_OB\x10\x05\x12\r\n\tDO_SPACES\x10\x06\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'proto.relics_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BUCKETMETADATA_BUCKETTAGSENTRY._options = None
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_options = b'8\001'
  _RELICFILE_HEADERSENTRY._options = None
  _RELICFILE_HEADERSENTRY._serialized_options = b'8\001'
  _RELICFILE_BODYENTRY._options = None
  _RELICFILE_BODYENTRY._serialized_options = b'8\001'
  _BACKEND._serialized_start=1148
  _BACKEND._serialized_end=1245
  _BACKENDINFO._serialized_start=22
  _BACKENDINFO._serialized_end=111
  _RELIC._serialized_start=114
  _RELIC._serialized_end=367
  _BUCKETMETADATA._serialized_start=370
  _BUCKETMETADATA._serialized_end=555
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_start=506
  _BUCKETMETADATA_BUCKETTAGSENTRY._serialized_end=555
  _RELICFILE._serialized_start=558
  _RELICFILE._serialized_end=1083
  _RELICFILE_HEADERSENTRY._serialized_start=935
  _RELICFILE_HEADERSENTRY._serialized_end=981
  _RELICFILE_BODYENTRY._serialized_start=983
  _RELICFILE_BODYENTRY._serialized_end=1026
  _RELICFILE_RELICTYPE._serialized_start=1028
  _RELICFILE_RELICTYPE._serialized_end=1083
  _RELICFILES._serialized_start=1085
  _RELICFILES._serialized_end=1146
# @@protoc_insertion_point(module_scope)


"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import abc
import grpc
import rsock_pb2
import typing

class RSockStub:
    def __init__(self, channel: grpc.Channel) -> None: ...
    Handshake: grpc.UnaryUnaryMultiCallable[
        rsock_pb2.HandshakeRequest,
        rsock_pb2.HandshakeResponse]

    Tunnel: grpc.StreamStreamMultiCallable[
        rsock_pb2.DataPacket,
        rsock_pb2.DataPacket]


class RSockServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Handshake(self,
        request: rsock_pb2.HandshakeRequest,
        context: grpc.ServicerContext,
    ) -> rsock_pb2.HandshakeResponse: ...

    @abc.abstractmethod
    def Tunnel(self,
        request_iterator: typing.Iterator[rsock_pb2.DataPacket],
        context: grpc.ServicerContext,
    ) -> typing.Iterator[rsock_pb2.DataPacket]: ...


def add_RSockServicer_to_server(servicer: RSockServicer, server: grpc.Server) -> None: ...
